     /////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM      //
   // Written by Andrew Broad                  ///
  //                                          ////
 // BinaryOpFrame.java (1998:11:24 17:47)    /////
//////////////////////////////////////////////////
// BinaryOpFrame is a subclass of Frame.     /////
//                                           /////
// Methods of BinaryOpFrame:                 /////
// - static initialise                       /////
// - BinaryOpFrame(BinaryOp,Frame)           /////
// - BinaryOpFrame(BinaryOp,Frame,Hashtable) /////
// - static buildQueuedFrames                /////
// - buildFrame                              /////
// - getFrameClass                           /////
// - getNumber             [INHERITED]       /////
// - setNumber             [INHERITED]       /////
// - getSlotFiller(Slot)   [INHERITED]       /////
// - getSlotFiller(String) [INHERITED]       /////
// - setSlotFiller(Slot)   [INHERITED]       /////
// - setSlotFiller(String) [INHERITED]       /////
// - toString              [INHERITED]       /////
// - print                 [INHERITED]       /////
// - generateFIL           [INHERITED]       ////
// - static writeClass                       ///
// - static readClass                        //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `BinaryOpFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava BinaryOp object - such frames
** should only be used to represent binary operations in EXPRESS models.
*******************************************************************************/

public class BinaryOpFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private BinaryOp _underlyingBinaryOp;
	private static Vector/*BinaryOpFrame*/ _toBeBuilt;
	private static Hashtable/*BinaryOp,BinaryOpFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static BinaryOpFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the BinaryOpFrame class by setting the class variable
	** _frameClass after looking up the `binary_operation' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `binary_operation'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("binary_operation");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY binary_operation in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// BinaryOpFrame constructor : BinaryOp ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BinaryOp object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the BinaryOp directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava BinaryOp object to wrap in this frame.
	** @param theEntity The entity which contains this binary operation.
 	** @exception FrameAlreadyExists If an BinaryOp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BinaryOpFrame (BinaryOp instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// BinaryOpFrame constructor : (BinaryOp x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BinaryOp object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the BinaryOp directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the binary operation frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava BinaryOp object to wrap in this frame.
	** @param theEntity The entity which contains this binary operation.
	** @param moreInstanceData Extra slots to tag this binary operation
	**  frame with.
	** @exception FrameAlreadyExists If an BinaryOp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BinaryOpFrame (BinaryOp instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingBinaryOp= instanceData;
		_theEntityOrType= theEntity;

		//////////////////////////////////////////////////////
		// Check that an BinaryOpFrame has not already been //
		// constructed for this BinaryOp.                   //
		//////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingBinaryOp))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingBinaryOp));

		_cache.put(_underlyingBinaryOp,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static BinaryOpFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any BinaryOp Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			BinaryOpFrame e= (BinaryOpFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// BinaryOpFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava BinaryOp
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			////////////////////////////////
			// operator : binary_operator //
			////////////////////////////////
			String enum= "";
			switch (_underlyingBinaryOp.kind) {
				case BinaryOp.LT: enum= "less_than"; break;
				case BinaryOp.LE: enum= "less_than_or_equals"; break;
				case BinaryOp.GT: enum= "greater_than";  break;
				case BinaryOp.GE: enum="greater_than_or_equals";break;
				case BinaryOp.EQ: enum= "equals"; break;
				case BinaryOp.NE: enum= "not_equals"; break;
				case BinaryOp.INST_EQ: enum= "instance_equals"; break;
				case BinaryOp.INST_NE: enum="instance_not_equals";
				                       break;
				case BinaryOp.IN: enum= "in_operator"; break;
				case BinaryOp.LIKE: enum= "like_operator"; break;
				case BinaryOp.ADD: enum= "add"; break;
				case BinaryOp.SUB: enum= "subtract"; break;
				case BinaryOp.OR: enum= "or_operator"; break;
				case BinaryOp.XOR: enum= "xor_operator"; break;
				case BinaryOp.MULT: enum= "multiply"; break;
				case BinaryOp.REAL_DIV: enum= "divide"; break;
				case BinaryOp.DIV: enum= "div_operator"; break;
				case BinaryOp.MOD: enum= "mod_operator"; break;
				case BinaryOp.AND: enum= "and_operator"; break;
				case BinaryOp.COMPLEX_CONSTRUCTOR:
					enum= "complex_constructor"; break;
				case BinaryOp.EXPONENT: enum= "exponent"; break;
			}
			setSlotFiller("operator", new EnumSymbol(enum));
	
			///////////////////////////////////
			// first_expression : expression //
			///////////////////////////////////
			setSlotFiller("first_expression", Values.convert
				(_underlyingBinaryOp.exp1, _theEntityOrType));

			////////////////////////////////////
			// second_expression : expression //
			////////////////////////////////////
			setSlotFiller("second_expression", Values.convert
				(_underlyingBinaryOp.exp2, _theEntityOrType));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "BinaryOpFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// BinaryOpFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static BinaryOpFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** BinaryOpFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static BinaryOpFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** BinaryOpFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
