     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM  //
   // Written by Andrew Broad              ///
  //                                      ////
 // BuiltInFrame.java (1998:10:27 11:00) /////
//////////////////////////////////////////////
// BuiltInFrame is a subclass of Frame.  /////
//                                       /////
// Methods of BuiltInFrame:              /////
// - static initialise                   /////
// - BuiltInFrame(BuiltIn)               /////
// - BuiltInFrame(BuiltIn,Hashtable)     /////
// - static buildQueuedFrames            /////
// - buildFrame                          /////
// - getFrameClass                       /////
// - getNumber             [INHERITED]   /////
// - setNumber             [INHERITED]   /////
// - getSlotFiller(Slot)   [INHERITED]   /////
// - getSlotFiller(String) [INHERITED]   /////
// - setSlotFiller(Slot)   [INHERITED]   /////
// - setSlotFiller(String) [INHERITED]   /////
// - toString              [INHERITED]   /////
// - print                 [INHERITED]   /////
// - generateFIL           [INHERITED]   ////
// - static writeClass                   ///
// - static readClass                    //
//////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `BuiltInFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava BuiltIn object - such frames
** should only be used to represent built-in constants in EXPRESS models.
*******************************************************************************/

public class BuiltInFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private BuiltIn _underlyingBuiltIn;
	private static Vector/*BuiltInFrame*/ _toBeBuilt;
	private static Hashtable/*BuiltIn,BuiltInFrame*/ _cache;
	
	/**********************************************************************/
	// static BuiltInFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the BuiltInFrame class by setting the class variable
	** _frameClass after looking up the `built_in_constant' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `built_in_constant'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("built_in_constant");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY built_in_constant in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// BuiltInFrame constructor : BuiltIn ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BuiltIn object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the BuiltIn directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava BuiltIn object to wrap in this frame.
	** @exception FrameAlreadyExists If an BuiltIn has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BuiltInFrame (BuiltIn instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// BuiltInFrame constructor : (BuiltIn x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava BuiltIn object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the BuiltIn directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the built-in constant frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava BuiltIn object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this built-in constant
	**  frame with.
	** @exception FrameAlreadyExists If an BuiltIn has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public BuiltInFrame (BuiltIn instanceData,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingBuiltIn= instanceData;

		/////////////////////////////////////////////////////
		// Check that an BuiltInFrame has not already been //
		// constructed for this BuiltIn.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingBuiltIn))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingBuiltIn));

		_cache.put(_underlyingBuiltIn,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static BuiltInFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any BuiltIn Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			BuiltInFrame e= (BuiltInFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// BuiltInFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava BuiltIn
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////////////////
			// built_in_value : built_in //
			///////////////////////////////
			String enum= "";
			switch (_underlyingBuiltIn.kind) {
				case BuiltIn.E:     enum= "e"; break;
				case BuiltIn.PI:    enum= "pie";   break;
				case BuiltIn.SELF:  enum= "this";  break;
				case BuiltIn.QMARK: enum= "infinity";  break;
			}
			setSlotFiller("built_in_value", new DefinedTypeInstance
			(DefinedType.lookup("built_in"),
			new EnumSymbol(enum)));
	
		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "BuiltInFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// BuiltInFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static BuiltInFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** BuiltInFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static BuiltInFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** BuiltInFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
