     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM ///
   // Written by Andrew Broad             ////
  //                                     /////
 // CaseLibrary.java (1998:10:27 11:03) //////
//////////////////////////////////////////////
// Methods of CaseLibrary:             ///////
// - CaseLibrary constructor           //////
// - CaseLibrary constructor(String)   /////
// - retrieve                          ////
// - insert                            ///
// - save                              //
////////////////////////////////////////

import java.io.*;
import java.util.*;

/*******************************************************************************
** The class `CaseLibrary' represents a case library. It encapsulates an
** organisation of cases within the library, allowing new cases to be added and
** existing cases to be retrieved.
*******************************************************************************/

public class CaseLibrary implements Serializable {
	private Hashtable/*FrameClass->Vector<Case>*/ _organ;

	/**********************************************************************/
	// CaseLibrary constructor : void ->
	/***********************************************************************
	** Constructs an instance of a case library which is initially empty.
	***********************************************************************/

	public CaseLibrary () {
		_organ= new Hashtable();
	}

	/**********************************************************************/
	// CaseLibrary constructor : String ->
	/***********************************************************************
	** Constructs an instance of a case library, loading it from a binary
	** file.
	** @param filename The filename of the case library binary.
	** @exception FileNotFoundException If the file does not exist.
	***********************************************************************/

	public CaseLibrary (String filename) throws FileNotFoundException {
		try {
			FileInputStream f= new FileInputStream(filename);
			ObjectInput s= new ObjectInputStream(f);
			_organ= (Hashtable)s.readObject();

			DefinedTypeFrame.readClass(s);
			EntityFrame.readClass(s);
			ExplicitAttrFrame.readClass(s);
			InverseAttrFrame.readClass(s);
			DerivedAttrFrame.readClass(s);
			DomainRuleFrame.readClass(s);
			UniqueRuleFrame.readClass(s);
			LiteralFrame.readClass(s);
			BuiltInFrame.readClass(s);
			UnaryOpFrame.readClass(s);
			BinaryOpFrame.readClass(s);
			QueryFrame.readClass(s);
			BuiltInFnCallFrame.readClass(s);
			UserFnCallFrame.readClass(s);
			QualifierFrame.readClass(s);
			AggInitFrame.readClass(s);
			AggInitElemFrame.readClass(s);
			AggTypeFrame.readClass(s);
			OneofFrame.readClass(s);
			AndFrame.readClass(s);
			AndorFrame.readClass(s);

			FrameClass.readClass(s);
			Slot.readClass(s);
			DefinedType.readClass(s);
		} catch (FileNotFoundException w) {
			_organ= new Hashtable();
			throw w;
		} catch (IOException x) {
			System.err.println(x);
			System.exit(1);
		} catch (ClassNotFoundException y) {
			System.err.println(y);
			System.exit(1);
		}		
	}

	/**********************************************************************/
	// CaseLibrary.retrieve : Frame -> Vector<Case>
	/***********************************************************************
	** Retrieves cases from the case library.
	** @param prompt A frame to use as a retrieval prompt.
	** @returns The set of cases which the prompt reminds the CBR system of.
	***********************************************************************/

	public Vector retrieve (Frame prompt) {
		Vector vector= (Vector)_organ.get(prompt.getFrameClass());
		if (vector == null)
			vector= new Vector();
		return vector;
	}

	/**********************************************************************/
	// CaseLibrary.insert : Case -> void
	/***********************************************************************
	** Adds a new case in the case library. Indexes it by the frame class
	** of the frame in its "problem" slot.
	** @param case The new case to store in the case library.
	***********************************************************************/

	public void insert (Case newCase) {
		try {
			FrameClass problemFrameClass= ((Frame)newCase.
			                              getCaseFrame().
		                                      getSlotFiller("problem")).
		                                      getFrameClass();
			Vector vector= (Vector)_organ.get(problemFrameClass);
			if (vector == null)
				_organ.put(problemFrameClass,
				           vector= new Vector());
			vector.addElement(newCase);
		} catch (InvalidSlot x) {
			System.err.println("Case does not have a \"problem\" "
			+ "slot!");
			newCase.getCaseFrame().print(System.err);
			System.exit(1);
		}
	}		

	/**********************************************************************/
	// CaseLibrary.save : String -> void
	/***********************************************************************
	** Saves the case library to a binary file.
	** @param filename The filename of the case library binary.
	***********************************************************************/

	public void save (String filename) {
		try {
			FileOutputStream f= new FileOutputStream(filename);
			ObjectOutput s= new ObjectOutputStream(f);
			s.writeObject(_organ);

			DefinedTypeFrame.writeClass(s);
			EntityFrame.writeClass(s);
			ExplicitAttrFrame.writeClass(s);
			InverseAttrFrame.writeClass(s);
			DerivedAttrFrame.writeClass(s);
			DomainRuleFrame.writeClass(s);
			UniqueRuleFrame.writeClass(s);
			LiteralFrame.writeClass(s);
			BuiltInFrame.writeClass(s);
			UnaryOpFrame.writeClass(s);
			BinaryOpFrame.writeClass(s);
			QueryFrame.writeClass(s);
			BuiltInFnCallFrame.writeClass(s);
			UserFnCallFrame.writeClass(s);
			QualifierFrame.writeClass(s);
			AggInitFrame.writeClass(s);
			AggInitElemFrame.writeClass(s);
			AggTypeFrame.writeClass(s);
			OneofFrame.writeClass(s);
			AndFrame.writeClass(s);
			AndorFrame.writeClass(s);

			FrameClass.writeClass(s);
			Slot.writeClass(s);
			DefinedType.writeClass(s);
		} catch (IOException x) {
			System.err.println(x);
			System.exit(1);
		}		
	}
}
