     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM ///
   // Written by Andrew Broad             ////
  //                                     /////
 // Constraints.java (1998:11:24 16:59) //////
/////////////////////////////////////////////
// Methods of Constraints:             /////
// - exjavaObjectsToFrames             ////
// - extraction                        ///
// - static main                       //
////////////////////////////////////////

import java.util.*;
import java.io.*;
import express.*;

/*******************************************************************************
** The class `Constraints' is the top-level class of the CONSTRAINTS
** system, with methods for converting ExJava objects to Frames,
** extracting higher-level constraints and transformation algorithms.
** It is not meant to be instantiated.
*******************************************************************************/

public class Constraints {
	private static CaseLibrary _caseLibrary;
	private static Vector/*Frame*/ _extractionQueue;
	private static Hashtable/*Frame->Boolean*/ _alreadyPrompted;
	private static Hashtable/*FrameClass->Vector<Frame>*/ _HLCsummary;

	private Constraints () {}

	/**********************************************************************/
	// static Constraints.exjavaObjectsToFrames : Express x boolean -> void
	/***********************************************************************
	** Converts the objects in the given ExJava EXPRESS model to
	** frames, just plugging them in as instances of their respective
	** frame classes. This method also creates frame classes from entities
	** and defined type classes from defined types just in case you want to
	** treat the model as a frame class model.
	** @param model The EXPRESS model.
	** @param createFrameClasses If true, a <a href="FrameClass.html">
	**  FrameClass</a> is created for each Entity in the model, and a
	**  <a href="DefinedType.html">DefinedType</a> for each type.
	** @returns A vector of the entity frames in the model.
	***********************************************************************/

	public static Vector exjavaObjectsToFrames (Express model,
	                                            boolean createFrameClasses)
	{
		Schema[] schemata= model.schemas;
		Vector entityFrames= new Vector();
		
		///////////////////////////////////////////////////////////////
		// For each exjava entity, create an entity frame            //
		// (causing all related frames to be created as needed) and, //
		// required, create a frame class while you're at it.        //
		///////////////////////////////////////////////////////////////
		for (int s= 0; s < schemata.length; s++) {
			System.out.println("\tSCHEMA " + schemata[s].id);
			Entity[] entities= schemata[s].entities;
			for (int e= 0; e < entities.length; e++) {
				try {
					entityFrames.addElement
						(new EntityFrame(entities[e]));
				} catch (FrameAlreadyExists x) {
					entityFrames.addElement(x._frame);
				}
				if (createFrameClasses)
					FrameClass.lookup(entities[e],model);
					//was new FrameClass(entities[e],model);
			}
			System.out.println("\tEND_SCHEMA " + schemata[s].id +
			                   "\n");
	        }
	        //////////////////////////////////////////////////////////////
	        // If required, create a defined type class for each exjava //
	        // defined type.                                            //
	        //////////////////////////////////////////////////////////////
	        if (createFrameClasses) {
			for (int s= 0; s < schemata.length; s++) {
				Type[] types= schemata[s].types;
				for (int t= 0; t < types.length; t++)
					new DefinedType(types[t]);
	       		}
		}
		return entityFrames;
	}
	
	/**********************************************************************/
	// static Constraints.extraction : Vector<EntityFrame> -> void
	/***********************************************************************
	** Extracts the higher level constraints from a set of entity frames
	** in an EXPRESS model.
	** @param entityFrames A vector of entity frames to extract HLCs from.
	***********************************************************************/

	public static void extraction (Vector entityFrames) {
		_alreadyPrompted= new Hashtable();
		_HLCsummary= new Hashtable();

		//////////////////////////////////////
		// Walk the model, extracting HLCs. //
		//////////////////////////////////////

		Enumeration entityEnumeration= entityFrames.elements();
		while (entityEnumeration.hasMoreElements()) {

			_extractionQueue= new Vector();

			EntityFrame entityFrame=
				(EntityFrame)entityEnumeration.nextElement();

			Frame conjunctionFrameForEntity= null;

			///////////////////////////////////////////////////////
			// Create a conjunction frame for the HLCs that will //
			// be extracted from this entity, and plant a link   //
			// to it from the entity frame. Also put the         //
			// conjunction frame on the extraction queue.        //
			///////////////////////////////////////////////////////
			try {
				conjunctionFrameForEntity= new Frame
					(FrameClass.lookup("conjunction"),
				        	new Hashtable());
				UnorderedVector entityFrameSingletonSet=
					new UnorderedVector();
				entityFrameSingletonSet.addElement(entityFrame);
				conjunctionFrameForEntity.setSlotFiller
					("extracted_from",
					 entityFrameSingletonSet);
				FrameClass fc= conjunctionFrameForEntity.
				               getFrameClass();
				Vector v= (Vector)_HLCsummary.get(fc);
				if (v == null) v= new Vector();
				v.addElement(conjunctionFrameForEntity);
				_HLCsummary.put(fc,v);
			} catch (FrameClassDoesNotExist x) {
				System.err.println("Frame class conjunction "
				+ "does not exist!");
				System.exit(1);
			} catch (InvalidSlot y) {
				System.err.println("Slot conjunction.extracted_"
				+ "from does not exist!");
				System.exit(1);
			}
			try {
				entityFrame.setSlotFiller
					("higher_level_constraints",
					conjunctionFrameForEntity);
			} catch (InvalidSlot z) {
				System.err.println("Slot entity_type.higher_"
				+ "level_constraints does not exist!");
				System.exit(1);
			}
			_extractionQueue.addElement(conjunctionFrameForEntity);

			////////////////////////////////////////////////////
			// Use the ENTITY itself as an extraction prompt. //
			////////////////////////////////////////////////////
			extractionPrompt(entityFrame,
			                 conjunctionFrameForEntity);

			////////////////////////////////////////////////////////
			// Use each EXPLICIT ATTRIBUTE as an extraction prompt//
			////////////////////////////////////////////////////////
			Vector explicitAttrFrames= null;
			try {
				explicitAttrFrames= (Vector)entityFrame.
					getSlotFiller("explicit_attributes");
			} catch (InvalidSlot x) {
				System.err.println("Invalid slot \"explicit_"
				+ "attributes\" in Constraints.extraction!");
				System.exit(1);
			}
			if (explicitAttrFrames != null) {
				Enumeration explicitAttrEnumeration=
					explicitAttrFrames.elements();
				while (explicitAttrEnumeration.
				       hasMoreElements())
				{
					extractionPrompt(
						(Frame)explicitAttrEnumeration.
						       nextElement(),
						conjunctionFrameForEntity);
				}
			}

			////////////////////////////////////////////////////////
			// Use each DERIVED ATTRIBUTE as an extraction prompt.//
			////////////////////////////////////////////////////////
			Vector derivedAttrFrames= null;
			try {
				derivedAttrFrames= (Vector)entityFrame.
					getSlotFiller("derived_attributes");
			} catch (InvalidSlot x) {
				System.err.println("Invalid slot \"derived_"
				+ "attributes\" in Constraints.extraction!");
				System.exit(1);
			}
			if (derivedAttrFrames != null) {
				Enumeration derivedAttrEnumeration=
					derivedAttrFrames.elements();
				while (derivedAttrEnumeration.
				       hasMoreElements())
				{
					extractionPrompt(
						(Frame)derivedAttrEnumeration.
						       nextElement(),
						conjunctionFrameForEntity);
				}
			}

			////////////////////////////////////////////////////////
			// Use each INVERSE ATTRIBUTE as an extraction prompt.//
			////////////////////////////////////////////////////////
			Vector inverseAttrFrames= null;
			try {
				inverseAttrFrames= (Vector)entityFrame.
					getSlotFiller("inverse_attributes");
			} catch (InvalidSlot x) {
				System.err.println("Invalid slot \"inverse_"
				+ "attributes\" in Constraints.extraction!");
				System.exit(1);
			}
			if (inverseAttrFrames != null) {
				Enumeration inverseAttrEnumeration=
					inverseAttrFrames.elements();
				while (inverseAttrEnumeration.
				       hasMoreElements())
				{
					extractionPrompt(
						(Frame)inverseAttrEnumeration.
						       nextElement(),
						conjunctionFrameForEntity);
				}
			}

			///////////////////////////////////////////////////
			// Use each UNIQUE RULE as an extraction prompt. //
			///////////////////////////////////////////////////
			Vector uniqueRuleFrames= null;
			try {
				uniqueRuleFrames= (Vector)entityFrame.
					getSlotFiller("unique_rules");
			} catch (InvalidSlot x) {
				System.err.println("Invalid slot \"unique_rules"
				+ "\" in Constraints.extraction!");
				System.exit(1);
			}
			if (uniqueRuleFrames != null) {
				Enumeration uniqueRuleEnumeration=
					uniqueRuleFrames.elements();
				while (uniqueRuleEnumeration.hasMoreElements())
				{
					extractionPrompt(
						(Frame)uniqueRuleEnumeration.
						       nextElement(),
						conjunctionFrameForEntity);
				}
			}

			//////////////////////////////////////////////////
			// Use the expression of each DOMAIN RULE as an //
			// extraction prompt.                           //
			//////////////////////////////////////////////////
			Vector domainRuleFrames= null;
			try {
				domainRuleFrames= (Vector)entityFrame.
					getSlotFiller("domain_rules");
			} catch (InvalidSlot x) {
				System.err.println("Invalid slot \"domain_rules"
				+ "\" in Constraints.extraction!");
				System.exit(1);
			}
			if (domainRuleFrames != null) {
				Enumeration domainRuleEnumeration=
					domainRuleFrames.elements();
				while (domainRuleEnumeration.hasMoreElements())
				{
					try {
						Frame expression= (Frame)
						((Frame)domainRuleEnumeration.
						nextElement()).getSlotFiller
						("expression");

						extractionPrompt(expression,
						conjunctionFrameForEntity);

					} catch (InvalidSlot y) {
						System.err.println("Invalid "
						+ "slot \"expression\" in "
						+ "Constraints.extraction!");
						System.exit(1);
					}
				}
			}

			//////////////////////////////////////////////
			// Walk the HLCs already extracted from the //
			// current entity, extracting further HLCs. //
			//////////////////////////////////////////////

			for (int i= 0; i < _extractionQueue.size(); i++) {
				extractionPrompt((Frame)_extractionQueue.
				                        elementAt(i), null);
			}
		}

		/////////////////////
		// Produce summary //
		/////////////////////

		System.out.println("----------------------------------------" +
			           "----------------------------------------");
		System.out.println("SUMMARY\n-------\n");
		System.out.println("The following higher-level " +
			"constraints have been extracted:");

		Enumeration HLCs= _HLCsummary.keys();
		while (HLCs.hasMoreElements()) {
			FrameClass key= (FrameClass)HLCs.nextElement();
			Vector v= (Vector)_HLCsummary.get(key);
			System.out.print("\n" + v.size() + " " + key + " HLC");
			if (v.size() > 1) System.out.print("s");
			System.out.println(":");
			
			Enumeration vHLCs= v.elements();
			while (vHLCs.hasMoreElements()) {
				System.out.println("\t" +
					((Frame)vHLCs.nextElement()));
			}
		}
	}

	/**********************************************************************/
	// static Constraints.extractionPrompt : (Frame x Frame) -> void
	/***********************************************************************
	** Extracts higher level constraints from a given prompt frame. Also
	** processes any subexpressions recursively.
	** @param promptFrame An prompt frame.
	** @param conjunctionFrame An optional CONJUNCTION frame on which to
	** hang any HLCs extracted from prompt frames in a conjunction.
	***********************************************************************/

	private static void extractionPrompt (Frame promptFrame,
	                                      Frame conjunctionFrame)
	{
		Frame conjunctionFrameForSubexpressions= null;

		/////////////////////////////////////////////////
		// Don't use any frame as an extraction prompt //
		// more than once.                             //
		/////////////////////////////////////////////////
		if (_alreadyPrompted.get(promptFrame) != null)
			return;
		else
			_alreadyPrompted.put(promptFrame, new Boolean(true));

		System.out.println("Submitting " + promptFrame + " as an "
		+ "extraction prompt.");
		
		//////////////////////////////////////////
		// Retrieve potentially matching cases. //
		//////////////////////////////////////////
		Vector cases= _caseLibrary.retrieve(promptFrame);

		////////////////////////////////////////////
		// Match cases to prompt, extracting HLCs //
		// when matches succeed.                  //
		////////////////////////////////////////////
		System.out.println("Matching it with the following cases...");
		
		Enumeration casesEnumeration= cases.elements();
		while (casesEnumeration.hasMoreElements()) {
			Case c= (Case)casesEnumeration.nextElement();

			System.out.print("\t" + c.getCaseFrame() + " - ");

			if (c.match(promptFrame)) {
				Frame HLC= c.extract();

				System.out.println("match succeeded, " + HLC
				+ " extracted.");

				_extractionQueue.addElement(HLC);
				FrameClass fc= HLC.getFrameClass();
				Vector v= (Vector)_HLCsummary.get(fc);
				if (v == null) v= new Vector();
				v.addElement(HLC);
				_HLCsummary.put(fc,v);

				 if (conjunctionFrame != null)
				 	try {
				 		HLC.setSlotFiller("conjunct_of",
				 	            conjunctionFrame);
				 	} catch (InvalidSlot x) {
				 		System.err.println("The slot " +
				 		"higher_level_constraint." +
				 		"conjunct_of does not exist!");
				 		System.exit(1);
				 	}

				 if (HLC.getFrameClass().getName().equals
				     ("conjunction"))
				 	conjunctionFrameForSubexpressions= HLC;
			} else
				System.out.println("match failed.");
		}
		System.out.println();


		////////////////////////////////
		////////////////////////////////
		//// PROCESS SUBEXPRESSIONS ////
		////////////////////////////////
		////////////////////////////////

		String kind= promptFrame.getFrameClass().getName();

		/////////////////////////////////////////////////
		// If the expression is a binary operation,    //
		// process its first and second subexpressions //
		/////////////////////////////////////////////////

		if (kind.equals("binary_operation")) {
			/////////////////////////
			// First subexpression //
			/////////////////////////
			try {
				Frame subexpression= (Frame)promptFrame.
					getSlotFiller("first_expression");

				extractionPrompt(subexpression,
					conjunctionFrameForSubexpressions);
			} catch (InvalidSlot x) {
				System.err.println("The slot \"first_"
				+ "expression\" was not found in the "
				+ "prompt frame:");
				promptFrame.print(System.err);
				System.exit(1);
			}
			//////////////////////////
			// Second subexpression //
			//////////////////////////
			try {
				Frame subexpression= (Frame)promptFrame.
				getSlotFiller("second_expression");

				extractionPrompt(subexpression,
					conjunctionFrameForSubexpressions);
			} catch (InvalidSlot y) {
				System.err.println("The slot \"second_"
				+ "expression\" was not found in the "
				+ "prompt frame:");
				promptFrame.print(System.err);
				System.exit(1);
			}
		}
	}

	/**********************************************************************/
	// Constraints.main : String -> void
	/***********************************************************************
	** The entry point to the CONSTRAINTS system.
	** @param argv The vector of command line arguments. There should be
	**  exactly two of these: the filename of an EXPRESS model (.exp) to
	**  parse using exjava, and a case library (.cas).
	***********************************************************************/

	public static void main (String argv[]) {
		try {
			String expressFilename= null;
			String caseLibraryFilename= null;

			if (argv.length != 2) {
				System.err.println("Usage: two arguments: "
				+ "an EXPRESS model (.exp file) and a case "
				+ "library (.cas file)");
				System.exit(1);
			}
			
			boolean parseFirst= false;

			for (int i= 0; i < 2; i++) {
				if (argv[i].endsWith(".exp")) {
					expressFilename= argv[i];
					parseFirst= false;
				}
				if (argv[i].endsWith(".cas")) {
					caseLibraryFilename= argv[i];
					parseFirst= true;
				}
			}

			if (expressFilename == null
			 || caseLibraryFilename == null)
			{
				System.err.println("Usage: two arguments: "
				+ "an EXPRESS model (.exp file) and a case "
				+ "library (.cas file)");
				System.exit(1);
			}
			Express inputModel= null;

			if (parseFirst) {
				System.out.println("Parsing EXPRESS model using"
				         + " ExJava...");
				inputModel= new Express(expressFilename);
				System.out.println
				           ("Done parsing EXPRESS model.\n");

				System.out.println("Loading case library...");
				_caseLibrary= new CaseLibrary
				                  (caseLibraryFilename);
				System.out.println
				           ("Done loading case library.\n");
			} else {
				System.out.println("Loading case library...");
				_caseLibrary= new CaseLibrary
				                  (caseLibraryFilename);
				System.out.println
				           ("Done loading case library.\n");

				System.out.println("Parsing EXPRESS model using"
				         + " ExJava...");
				inputModel= new Express(expressFilename);
				System.out.println
				           ("Done parsing EXPRESS model.\n");
			}

			System.out.println("Converting ExJava objects to frames"
			                 + "...");
	 		Vector entityFrames= exjavaObjectsToFrames
	 		                     (inputModel,false);
	 		System.out.println("Done converting ExJava objects to "
	 		                 + "frames.\n");

			extraction(entityFrames);

			Frame.generateFIL(true);
	        } catch (Exception x) {
	        	System.err.println(
	        		"----------------------------------------" +
	        		"----------------------------------------");
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "Constraints.main!\n\ttoString == " + x + "\n");
			System.err.println("Its Message is:-\n" + x.getMessage()
			+ "\n");
			System.err.println("Its LocalizedMessage is:-\n"
			+ x.getLocalizedMessage());
			System.err.println("Printing stack trace to " +
			"\"StackTrace.txt\"...\n");
			try {
				x.printStackTrace(new PrintWriter
				(new FileOutputStream("StackTrace.txt"),true));
			} catch (IOException y) {
				System.err.println("IOException");
			}
	        }  
	}
}
