     /////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM ///
   // Written by Andrew Broad             ////
  //                                     /////
 // DefinedType.java (1998:10:27 11:06) //////
//////////////////////////////////////////////
// Methods of DefinedType:             ///////
// - DefinedType constructor           ///////
// - getTypeFrame                      ///////
// - validateValue                     ///////
// - toString                          ///////
// - static getAllFrameClasses         //////
// - static lookup(String)             /////
// - static lookup(String,Express)     ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `DefinedType' represents a defined type which can have instances
** in the frame system, doing for <a href="DefinedTypeInstance.html">
** DefinedTypeInstance</a> what <a href="FrameClass.html">FrameClass</a> does
** for <a href="Frame.html">Frame</a>.
*******************************************************************************/

public class DefinedType implements Serializable {
	private Frame _typeFrame;
	private static Hashtable/*String,DefinedType*/ _definedTypeIndex;

        /**********************************************************************/
        // DefinedType constructor : Type ->
        /***********************************************************************
        ** Constructs this defined type by plugging in the given exjava Type
        ** object which is taken to represent it (it does so by copying the
        ** reference, which is maximally efficient, but the client should not
        ** then alter the Type directly, as that would compromise the
        ** integrity of the frame).
        ** @param typeFrame An exjava Type object to wrap in this frame.
       ***********************************************************************/

	public DefinedType (Type typeFrame) {
		if (_definedTypeIndex == null)
			_definedTypeIndex= new Hashtable();
					
		try {
			_typeFrame= new DefinedTypeFrame(typeFrame);
		} catch (FrameAlreadyExists y) {
			_typeFrame= y._frame;
		}

		_definedTypeIndex.put(typeFrame.id,this);
	}

	/**********************************************************************/
	// DefinedType.getTypeFrame : void -> Frame
	/***********************************************************************
	** Delivers up the underlying frame of this defined type.
	** @return The class frame representing this defined type.
	***********************************************************************/
 
	public Frame getTypeFrame () {
		return _typeFrame;
	}
	
	/**********************************************************************/
	// DefinedType.validateValue : Object -> void
	/***********************************************************************
	** Checks that the given value conforms to the underlying type of this
	** defined type, raising exceptions if it does not. Validation is not
	** actually implemented - this is just a hook.
	** @param value The value to validate.
	***********************************************************************/
 
 	public void validateValue (Object value)
	                          /* throws InvalidTypeValue */
	{}
	
	/**********************************************************************/
	// DefinedType.toString : void -> String
	/***********************************************************************
	** Returns the name of this defined type.
	** @returns The string representation of this defined type.
	***********************************************************************/

	public String toString () {
		try {
			return (String)_typeFrame.getSlotFiller("name");
		} catch (InvalidSlot x) {
			System.err.println("Something has gone "
			+ "horribly wrong for DefinedType."
			+ "toString to throw an InvalidSlot!");
			System.exit(1); return "";
		}
	}

        /**********************************************************************/
        // static DefinedType.getAllDefinedTypes : void -> Enumeration
        /***********************************************************************
        ** Returns all the defined types in the frame system.
        ** @return An enumeration of all the defined types in the frame system.
        ***********************************************************************/

	public static Enumeration getAllDefinedTypes () {
		return _definedTypeIndex.elements();
	}

        /**********************************************************************/
        // static DefinedType.lookup : String -> DefinedType
        /***********************************************************************
        ** Looks up the defined type with the given name.
        ** @param definedTypeName The name of the frame class to return.
        ** @return The FrameClass object whose name is `frameClassName'.
        ** @exception DefinedTypeDoesNotExist If there is not a frame class of
        **  that name.
        ***********************************************************************/
 
        public static DefinedType lookup (String definedTypeName)
                            throws DefinedTypeDoesNotExist
        {
		DefinedType definedType= (DefinedType)_definedTypeIndex.
			get(definedTypeName);
		if (definedType == null)
			throw new DefinedTypeDoesNotExist();

        	return definedType;
        }

        /**********************************************************************/
        // static DefinedType.lookup : String x Express -> DefinedType
        /***********************************************************************
        ** Looks up the defined type with the given name.
        ** @param definedTypeName The name of the defined type to return.
        ** @param frameClassModel The EXPRESS model in which `definedTypeName'
        **  is defined, in case that defined type hasn't been constructed yet.
        ** @return The DefinedType object whose name is `definedTypeName'.
        ** @exception DefinedTypeDoesNotExist If there is not a defined type of
        **  that name.
        ***********************************************************************/
 
        public static DefinedType lookup (String definedTypeName,
                                         Express frameClassModel)
                            throws DefinedTypeDoesNotExist
        {
        	DefinedType definedType= (DefinedType)_definedTypeIndex.
        		get(definedTypeName);
        	if (definedType == null) {
        		Type t= frameClassModel.findType(definedTypeName);
        		if (t != null)
        			definedType= new DefinedType(t);
        		else
        			throw new DefinedTypeDoesNotExist();
        	}
        	return definedType;
        }

	/**********************************************************************/
	// static DefinedType.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of DefinedType.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_definedTypeIndex);
	}

	/**********************************************************************/
	// static DefinedType.readClass : ObjectInputStream -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** DefinedType.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_definedTypeIndex= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the"
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
