     /////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM      //
   // Written by Andrew Broad                  ///
  //                                          ////
 // DerivedAttrFrame.java (1998:10:27 11:09) /////
//////////////////////////////////////////////////
// DerivedAttrFrame is a subclass of Frame.  /////
//                                           /////
// Methods of DerivedAttrFrame:              /////
// - static initialise                       /////
// - DerivedAttrFrame(DerivedAttr)           /////
// - DerivedAttrFrame(DerivedAttr,Hashtable) /////
// - static buildQueuedFrames                /////
// - buildFrame                              /////
// - getFrameClass                           /////
// - getNumber             [INHERITED]       /////
// - setNumber             [INHERITED]       /////
// - getSlotFiller(Slot)   [INHERITED]       /////
// - getSlotFiller(String) [INHERITED]       /////
// - setSlotFiller(Slot)   [INHERITED]       /////
// - setSlotFiller(String) [INHERITED]       /////
// - toString              [INHERITED]       /////
// - print                 [INHERITED]       /////
// - generateFIL           [INHERITED]       ////
// - static writeClass                       ///
// - static readClass                        //
//////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `DerivedAttrFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava DerivedAttr object - such frames
** should only be used to represent derived attributes in EXPRESS models.
*******************************************************************************/

public class DerivedAttrFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private DerivedAttr _underlyingDerivedAttr;
	private static Vector/*DerivedAttrFrame*/ _toBeBuilt;
	private static Hashtable/*DerivedAttr,DerivedAttrFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static DerivedAttrFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the DerivedAttrFrame class by setting the class variable
	** _frameClass after looking up the `derived_attribute' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `derived_attribute'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("derived_attribute");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY derived_attribute in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// DerivedAttrFrame constructor : DerivedAttr ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava DerivedAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the DerivedAttr
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava DerivedAttr object to wrap in this
	**                     frame.
	** @param theEntity The entity which contains this derived attribute.
	** @exception FrameAlreadyExists If an DerivedAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public DerivedAttrFrame (DerivedAttr instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// DerivedAttrFrame constructor :
	//	(DerivedAttr x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava DerivedAttr
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the DerivedAttr
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the derived attribute frame to be tagged with
	** extra slots, given in a hash table.
	** @param instanceData An exjava DerivedAttr object to wrap in this
	**                     frame.
	** @param moreInstanceData Extra slots to tag this derived attribute
	**  frame with.
	** @param theEntity The entity which contains this derived attribute.
	** @exception FrameAlreadyExists If an DerivedAttrFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public DerivedAttrFrame (DerivedAttr instanceData, Frame theEntity,
	                         Hashtable moreInstanceData)
	                         throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingDerivedAttr= instanceData;
		_theEntityOrType= theEntity;

		/////////////////////////////////////////////////////////
		// Check that an DerivedAttrFrame has not already been //
		// constructed for this DerivedAttr.                   //
		/////////////////////////////////////////////////////////

		if (_cache.containsKey(_underlyingDerivedAttr))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingDerivedAttr));

		_cache.put(_underlyingDerivedAttr,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static DerivedAttrFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any DerivedAttrFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			DerivedAttrFrame e= (DerivedAttrFrame)
			                    toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// DerivedAttrFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** DerivedAttr object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////
			// name : STRING //
			///////////////////
			setSlotFiller("name",_underlyingDerivedAttr.id);

			///////////////////////////
			// tipe : parameter_type //
			///////////////////////////
			setSlotFiller("tipe", Values.convert
				(_underlyingDerivedAttr.type));

			//////////////////////////////////
			// owner : OPTIONAL entity_type //
			//////////////////////////////////
			setSlotFiller("owner", Values.convert
				(_underlyingDerivedAttr.owner));
			
			//////////////////////////
			// formula : expression //
			//////////////////////////
			setSlotFiller("formula", Values.convert
				(_underlyingDerivedAttr.exp, _theEntityOrType));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "DerivedAttrFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// DerivedAttrFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static DerivedAttrFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** DerivedAttrFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static DerivedAttrFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** DerivedAttrFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
