     /////////////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM          //
   // Written by Andrew Broad                      ///
  //                                              ////
 // DomainRuleFrame.java (1998:10:27 11:09)      /////
//////////////////////////////////////////////////////
// DomainRuleFrame is a subclass of Frame.       /////
//                                               /////
// Methods of DomainRuleFrame:                   /////
// - static initialise                           /////
// - DomainRuleFrame(DomainRule,Frame)           /////
// - DomainRuleFrame(DomainRule,Frame,Hashtable) /////
// - static buildQueuedFrames                    /////
// - buildFrame                                  /////
// - getFrameClass                               /////
// - getNumber             [INHERITED]           /////
// - setNumber             [INHERITED]           /////
// - getSlotFiller(Slot)   [INHERITED]           /////
// - getSlotFiller(String) [INHERITED]           /////
// - setSlotFiller(Slot)   [INHERITED]           /////
// - setSlotFiller(String) [INHERITED]           /////
// - toString              [INHERITED]           /////
// - print                 [INHERITED]           /////
// - generateFIL           [INHERITED]           ////
// - static writeClass                           ///
// - static readClass                            //
//////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `DomainRuleFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava DomainRule object - such frames
** should only be used to represent domain rules in EXPRESS models.
*******************************************************************************/

public class DomainRuleFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private DomainRule _underlyingDomainRule;
	private static Vector/*DomainRuleFrame*/ _toBeBuilt;
	private static Hashtable/*DomainRule,DomainRuleFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static DomainRuleFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the DomainRuleFrame class by setting the class variable
	** _frameClass after looking up the `domain_rule' entity in the given
	** EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `domain_rule'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("domain_rule");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY domain_rule in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// DomainRuleFrame constructor : DomainRule ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava DomainRule
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the DomainRule
	** directly, as that would compromise the integrity of the frame).
	** @param instanceData An exjava DomainRule object to wrap in this
	**                     frame.
	** @param theEntity The entity which contains this domain rule.
	** @exception FrameAlreadyExists If an DomainRuleFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public DomainRuleFrame (DomainRule instanceData, Frame theEntity)
	                       throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// DomainRuleFrame constructor : (DomainRule x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava DomainRule
	** object (it does so by copying the reference, which is maximally
	** efficient, but the client should not then alter the DomainRule
	** directly, as that would compromise the integrity of the frame). This
	** constructor allows the domain rule frame to be tagged with
	** extra slots, given in a hash table.
	** @param instanceData An exjava DomainRule object to wrap in this
	**                     frame.
	** @param theEntity The entity which contains this domain rule.
	** @param moreInstanceData Extra slots to tag this domain rule frame
	**  with.
	** @exception FrameAlreadyExists If an DomainRuleFrame has already
	**  been constructed for `instanceData'.
	***********************************************************************/

	public DomainRuleFrame (DomainRule instanceData, Frame theEntity,
	                        Hashtable moreInstanceData)
	                        throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingDomainRule= instanceData;
		_theEntityOrType= theEntity;

		////////////////////////////////////////////////////////
		// Check that an DomainRuleFrame has not already been //
		// constructed for this DomainRule.                   //
		////////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingDomainRule))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingDomainRule));

		_cache.put(_underlyingDomainRule,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
					+ "wrong for an InvalidFrameClass to be"
						+ " thrown here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static DomainRuleFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any DomainRuleFrames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			DomainRuleFrame e= (DomainRuleFrame)
				           toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// DomainRuleFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava
	** DomainRule object.
	***********************************************************************/

	public void buildFrame () {
		try {
			////////////////////
			// label : STRING //
			////////////////////
			setSlotFiller("label",_underlyingDomainRule.id);
		
			/////////////////////////////
			// expression : expression //
			/////////////////////////////
			setSlotFiller("expression", Values.convert
				(_underlyingDomainRule.exp, _theEntityOrType));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "DomainRuleFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// DomainRuleFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static DomainRuleFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** DomainRuleFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static DomainRuleFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** DomainRuleFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
