     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // EntityFrame.java (1998:11:03 10:58) /////
/////////////////////////////////////////////
// EntityFrame is a subclass of Frame.  /////
//                                      /////
// Methods of EntityFrame:              /////
// - static initialise                  /////
// - EntityFrame(Entity)                /////
// - EntityFrame(Entity,Hashtable)      /////
// - static buildQueuedFrames           /////
// - buildFrame                         /////
// - getFrameClass                      /////
// - getNumber             [INHERITED]  /////
// - setNumber             [INHERITED]  /////
// - getSlotFiller(Slot)   [INHERITED]  /////
// - getSlotFiller(String) [INHERITED]  /////
// - setSlotFiller(Slot)   [INHERITED]  /////
// - setSlotFiller(String) [INHERITED]  /////
// - toString              [INHERITED]  /////
// - print                 [INHERITED]  /////
// - generateFIL           [INHERITED]  /////
// - static writeClass                  /////
// - static readClass                   /////
// - lookupAttribute                    /////
// - getQueryVariables                  ////
// - pushQueryVariable                  ///
// - popQueryVariable                   //
/////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `EntityFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava Entity object - such frames should
** only be used to represent entities in EXPRESS models.
*******************************************************************************/

public class EntityFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private Entity _underlyingEntity;
	private static Vector/*EntityFrame*/ _toBeBuilt;
	private static Hashtable/*Entity->EntityFrame*/ _cache;
	private Hashtable/*String->AttrFrame*/ _attrCache;
	private Vector/*Frame<query_variable>*/ _queryVariableStack;

	/**********************************************************************/
	// static EntityFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the EntityFrame class by setting the class variable
	** _frameClass after looking up the `entity_type' entity in the given
	** EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `entity_type'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("entity_type");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY entity_type in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}

	/**********************************************************************/
	// EntityFrame constructor : Entity ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Entity object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the Entity directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava Entity object to wrap in this frame.
	** @exception FrameAlreadyExists If an EntityFrame has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public EntityFrame (Entity instanceData)
	                   throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}

	/**********************************************************************/
	// EntityFrame constructor : (Entity x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Entity object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the Entity directly, as that
	** would compromise the integrity of the frame). This constructor
	** allows the entity frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava Entity object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this entity frame with.
	** @exception FrameAlreadyExists If an EntityFrame has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public EntityFrame (Entity instanceData,
	                    Hashtable moreInstanceData)
	                   throws FrameAlreadyExists
	{
		super(moreInstanceData);
		
		if (_cache == null)
			_cache= new Hashtable();

		_underlyingEntity= instanceData;

		_queryVariableStack= new Vector();

		////////////////////////////////////////////////////////////////
		// Check that an EntityFrame has not already been constructed //
		// for this Entity.                                           //
		////////////////////////////////////////////////////////////////
		if (_cache.containsKey(_underlyingEntity))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingEntity));

//System.err.println("\t\tENTITY " + _underlyingEntity.id);

		_cache.put(_underlyingEntity,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
 			buildFrame();
			_frameClass.validateFrame(this);
 			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
              		setNumber(_frameClass.getInstances().size());

/*			try {
//System.out.println("\t\tEND_ENTITY " + getSlotFiller("name") + "\n");
			} catch (InvalidSlot y) {
				System.err.println("EntityFrame.buildQueued" +
				"Frames: InvalidSlot");
				System.exit(1);
			}
*/		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}

	/**********************************************************************/
	// static EntityFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any EntityFrames that are queued to be built (those that were
	** constructed before the frame class was constructed). Their slots are
	** filled from the exjava objects with which they were constructed, and
	** they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			EntityFrame e= (EntityFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());

/*			try {
//System.err.println("\t\tEND_ENTITY " + e.getSlotFiller("name") + "\n");
			} catch (InvalidSlot y) {
				System.err.println("EntityFrame.buildQueued" +
				"Frames: InvalidSlot");
				System.exit(1);
			}
*/		}
	}

	/**********************************************************************/
	// EntityFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava Entity
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
		        ///////////////////
			// name : STRING //
			///////////////////
			setSlotFiller("name",_underlyingEntity.id);
			
			///////////////////////////
			// is_abstract : BOOLEAN //
			///////////////////////////
			setSlotFiller("is_abstract",new Logical
				(_underlyingEntity.isAbstract
				                  ? Logical.TRUE
				                  : Logical.FALSE));
				 
			//////////////////////////////////////////////////
			// supertype_of : OPTIONAL supertype_expression //
			//////////////////////////////////////////////////
			setSlotFiller("supertype_of", Values.convert
				(_underlyingEntity.supertype));

			//////////////////////////////////////////////////////
			// subtype_of : OPTIONAL SET [1 : ?] OF entity_type //
			//////////////////////////////////////////////////////
			if (_underlyingEntity.subtype.length > 0)
				setSlotFiller("subtype_of", new UnorderedVector
					(_underlyingEntity.subtype));

			////////////////////////////////////////////
			// explicit_attributes :                  //
			//      SET [0 : ?] OF explicit_attribute //
			////////////////////////////////////////////
			setSlotFiller("explicit_attributes", new
				UnorderedVector(_underlyingEntity.explicits));

			////////////////////////////////////////////////////
			// derived_attributes :                           //
			//      OPTIONAL SET [1 : ?] OF derived_attribute //
			////////////////////////////////////////////////////
			if (_underlyingEntity.deriveds.length > 0)
				setSlotFiller("derived_attributes", new
				UnorderedVector(_underlyingEntity.deriveds,
				                this));

			////////////////////////////////////////////////////
			// inverse_attributes :                           //
			//      OPTIONAL SET [1 : ?] OF inverse_attribute //
			////////////////////////////////////////////////////
			if (_underlyingEntity.inverses.length > 0)
				setSlotFiller("inverse_attributes", new
				UnorderedVector(_underlyingEntity.inverses));

			//////////////////////////////////////////////
			// unique_rules :                           //
			//      OPTIONAL SET [1 : ?] OF unique_rule //
			//////////////////////////////////////////////
			if (_underlyingEntity.uniques.length > 0)
				setSlotFiller("unique_rules", new
				UnorderedVector(_underlyingEntity.uniques));

			////////////////////////////////////////////////////////
			// domain_rules : OPTIONAL SET [1 : ?] OF domain_rule //
			////////////////////////////////////////////////////////
			if (_underlyingEntity.rules.length > 0)
				setSlotFiller("domain_rules", new
				UnorderedVector(_underlyingEntity.rules,this));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "EntityFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// EntityFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}

	/**********************************************************************/
	// static EntityFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of EntityFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static EntityFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** EntityFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}

        /**********************************************************************/
        // EntityFrame.lookupAttribute : String -> Frame
        /***********************************************************************
        ** Looks up the attribute with the given name.
        ** @param attrName The name of the attribute to return.
        ** @return The attribute frame whose name is `attrName'.
        ** @exception InvalidAttribute If the entity frame does not have an
        **                             attribute of that name.
        ***********************************************************************/
 
	public Frame lookupAttribute (String attrName) throws InvalidAttribute {
		if (_attrCache == null)
			_attrCache= new Hashtable();
		Frame attribute= (Frame)_attrCache.get(attrName);
		if (attribute != null)
			return attribute;
		try {
		///////////////////////////////////////////////////////////
		// If the attribute isn't in the cache already,          //
		// look for it in the explicit attributes of this entity //
		///////////////////////////////////////////////////////////
			Enumeration explicitAttributes= ((UnorderedVector)
				getSlotFiller("explicit_attributes")).
				elements();
			while (explicitAttributes.hasMoreElements()) {
				Frame attr= (Frame)explicitAttributes.
				                   nextElement();
				if(((String)attr.getSlotFiller("name")).
				    equals(attrName))
					return attr;
			}
		//////////////////////////////////////////////////////////
		// look for it in the inverse attributes of this entity //
		//////////////////////////////////////////////////////////
			Enumeration inverseAttributes= ((UnorderedVector)
				getSlotFiller("inverse_attributes")).
					elements();
			while (inverseAttributes.hasMoreElements()) {
				Frame attr= (Frame)inverseAttributes.
				    nextElement();
				if (((String)attr.getSlotFiller("name")).
					     equals(attrName))
				    return attr;
			}
		//////////////////////////////////////////////////////////
		// look for it in the derived attributes of this entity //
		//////////////////////////////////////////////////////////
			Enumeration derivedAttributes= ((UnorderedVector)
				getSlotFiller("derived_attributes")).
					elements();
			while (derivedAttributes.hasMoreElements()) {
				Frame attr= (Frame)derivedAttributes.
				    nextElement();
				if (((String)attr.getSlotFiller("name")).
					     equals(attrName))
				    return attr;
			}
		///////////////////////////////////////////////////
		// look it up in the supertype(s) of this entity //
		///////////////////////////////////////////////////
			Enumeration supertypes= ((UnorderedVector)
				getSlotFiller("subtype_of")).elements();
			while (supertypes.hasMoreElements()) {
				try {
					Frame attr= ((EntityFrame)supertypes.
					    nextElement()).
						lookupAttribute(attrName);
					return attr;
				} catch (InvalidAttribute x) {}
			}
		} catch (InvalidSlot x) {
			System.err.println("Something's gone horribly wrong "
			+ "for an InvalidSlot to be thrown in EntityFrame."
			+ "lookupAttribute!");
			System.exit(1);
		} catch (NullPointerException e) {
			//System.err.println("Warning: EntityFrame."
			//+ "lookupAttribute: bailing out because the entity "
			//+ "frame is not complete yet:");
			//print(System.err);
		}
		//////////////////////////////////////////////
		// If after all that you fail to find it... //
		//////////////////////////////////////////////
		throw new InvalidAttribute();
	}

	/**********************************************************************/
	// EntityFrame.getQueryVariables : void -> Vector
	/***********************************************************************
	** Returns the stack of query variables on this frame.
	** @return The query variables on this entity frame.
	***********************************************************************/
 
	public Vector getQueryVariables () {
		return _queryVariableStack;
	}

	/**********************************************************************/
	// EntityFrame.pushQueryVariable : Frame -> void
	/***********************************************************************
	** Pushes a given query variable onto the stack.
	** @param queryVariable The query variable to be pushed onto this frame,
	**                      which is supposed to be a query_variable frame.
	***********************************************************************/
 
	public void pushQueryVariable (Frame queryVariable) {
		_queryVariableStack.addElement(queryVariable);
	}

	/**********************************************************************/
	// EntityFrame.popQueryVariable : void -> void
	/***********************************************************************
	** Pops the top query variable off of this frame.
	***********************************************************************/
 
	public void popQueryVariable () {
		_queryVariableStack.removeElementAt(0);
	}
}
