     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // OneofFrame.java (1998:10:27 11:32)  /////
/////////////////////////////////////////////
// OneofFrame is a subclass of Frame.   /////
//                                      /////
// Methods of OneofFrame:               /////
// - static initialise                  /////
// - OneofFrame(SupExp,Frame)           /////
// - OneofFrame(SupExp,Frame,Hashtable) /////
// - static buildQueuedFrames           /////
// - buildFrame                         /////
// - getFrameClass                      /////
// - getNumber             [INHERITED]  /////
// - setNumber             [INHERITED]  /////
// - getSlotFiller(Slot)   [INHERITED]  /////
// - getSlotFiller(String) [INHERITED]  /////
// - setSlotFiller(Slot)   [INHERITED]  /////
// - setSlotFiller(String) [INHERITED]  /////
// - toString              [INHERITED]  /////
// - print                 [INHERITED]  /////
// - generateFIL           [INHERITED]  ////
// - static writeClass                  ///
// - static readClass                   //
/////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `OneofFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava SupExp object - such frames
** should only be used to represent ONEOF operations in EXPRESS models.
*******************************************************************************/

public class OneofFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private SupExp _underlyingSupExp;
	private static Vector/*OneofFrame*/ _toBeBuilt;
	private static Hashtable/*SupExp,OneofFrame*/ _cache;
	
	/**********************************************************************/
	// static OneofFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the OneofFrame class by setting the class variable
	** _frameClass after looking up the `oneof_expression' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `oneof_expression'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("oneof_expression");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY oneof_expression in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// OneofFrame constructor : SupExp ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava SupExp object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the SupExp directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava SupExp object to wrap in this frame.
 	** @exception FrameAlreadyExists If an SupExp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public OneofFrame (SupExp instanceData)
	                         throws FrameAlreadyExists
	{
		this(instanceData,new Hashtable());
	}
	
	/**********************************************************************/
	// OneofFrame constructor : (SupExp x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava SupExp object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the SupExp directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the ONEOF operation frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava SupExp object to wrap in this frame.
	** @param moreInstanceData Extra slots to tag this ONEOF operation
	**  frame with.
	** @exception FrameAlreadyExists If an SupExp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public OneofFrame (SupExp instanceData, Hashtable moreInstanceData)
	                  throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingSupExp= instanceData;

		/////////////////////////////////////////////////////
		// Check that an OneofFrame has not already been //
		// constructed for this SupExp.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingSupExp))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingSupExp));

		_cache.put(_underlyingSupExp,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static OneofFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any SupExp Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			OneofFrame e= (OneofFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// OneofFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava SupExp
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////////////////////////////////////////
			// expressions : SET [1 : ?] OF supertype_expression //
			///////////////////////////////////////////////////////
			setSlotFiller("expressions", new UnorderedVector
				(_underlyingSupExp.exps));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "OneofFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// OneofFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static OneofFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** OneofFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static OneofFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** OneofFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
