     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // QueryFrame.java (1998:10:27 11:34)  /////
/////////////////////////////////////////////
// QueryFrame is a subclass of Frame.   /////
//                                      /////
// Methods of QueryFrame:               /////
// - static initialise                  /////
// - QueryFrame(Query,Frame)            /////
// - QueryFrame(Query,Frame,Hashtable)  /////
// - static buildQueuedFrames           /////
// - buildFrame                         /////
// - getFrameClass                      /////
// - getNumber             [INHERITED]  /////
// - setNumber             [INHERITED]  /////
// - getSlotFiller(Slot)   [INHERITED]  /////
// - getSlotFiller(String) [INHERITED]  /////
// - setSlotFiller(Slot)   [INHERITED]  /////
// - setSlotFiller(String) [INHERITED]  /////
// - toString              [INHERITED]  /////
// - print                 [INHERITED]  /////
// - generateFIL           [INHERITED]  ////
// - static writeClass                  ///
// - static readClass                   //
/////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `QueryFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava Query object - such frames
** should only be used to represent queries in EXPRESS models.
*******************************************************************************/

public class QueryFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private Query _underlyingQuery;
	private static Vector/*QueryFrame*/ _toBeBuilt;
	private static Hashtable/*Query,QueryFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static QueryFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the QueryFrame class by setting the class variable
	** _frameClass after looking up the `query_expression' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `query_expression'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("query_expression");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY query_expression in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// QueryFrame constructor : Query ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Query object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the Query directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava Query object to wrap in this frame.
	** @param theEntity The entity which contains this binary operation.
 	** @exception FrameAlreadyExists If an Query has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public QueryFrame (Query instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// QueryFrame constructor : (Query x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava Query object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the Query directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the binary operation frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava Query object to wrap in this frame.
	** @param theEntity The entity which contains this binary operation.
	** @param moreInstanceData Extra slots to tag this binary operation
	**  frame with.
	** @exception FrameAlreadyExists If an Query has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public QueryFrame (Query instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingQuery= instanceData;
		_theEntityOrType= theEntity;

		/////////////////////////////////////////////////////
		// Check that an QueryFrame has not already been //
		// constructed for this Query.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingQuery))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingQuery));

		_cache.put(_underlyingQuery,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static QueryFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any Query Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			QueryFrame e= (QueryFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// QueryFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava Query
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////////////////
			// variable : query_variable //
			///////////////////////////////
			Frame qVFrame= null;
			try {
				qVFrame= new Frame(FrameClass.
			                           lookup("query_variable"),
			                           new Hashtable());
			} catch (FrameClassDoesNotExist x) {
				System.err.println("The frame class \"query_" +
				"variable\" does not exist!");
				System.exit(1);
			}
			try {
				qVFrame.setSlotFiller("name",
				                      _underlyingQuery.id);
			} catch (InvalidSlot y) {
				System.err.println("The slot \"query_variable."
				+ "name\" does not exist!");
				System.exit(1);
			}

			setSlotFiller("variable", qVFrame);

			if (_theEntityOrType instanceof EntityFrame)
				((EntityFrame)_theEntityOrType).
					pushQueryVariable(qVFrame);
			else if (_theEntityOrType instanceof DefinedTypeFrame)
				((DefinedTypeFrame)_theEntityOrType).
					pushQueryVariable(qVFrame);
			else {
				System.err.println("QueryFrame.buildFrame: " +
				"_theEntityOrTypeFrame is not either an entity_"
				+ "type or a defined_type frame!");
				System.exit(1);
			}
	
			/////////////////////////////////////////
			// aggregation_expression : expression //
			/////////////////////////////////////////
			setSlotFiller("aggregation_expression", Values.convert
				(_underlyingQuery.agg, _theEntityOrType));

			/////////////////////////////////////
			// logical_expression : expression //
			/////////////////////////////////////
			setSlotFiller("logical_expression", Values.convert
				(_underlyingQuery.exp, _theEntityOrType));

			if (_theEntityOrType instanceof EntityFrame)
				((EntityFrame)_theEntityOrType).
					popQueryVariable();
			else if (_theEntityOrType instanceof DefinedTypeFrame)
				((DefinedTypeFrame)_theEntityOrType).
					popQueryVariable();
			else {
				System.err.println("QueryFrame.buildFrame: " +
				"_theEntityOrTypeFrame is not either an entity_"
				+ "type or a defined_type frame!");
				System.exit(1);
			}
		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "QueryFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// QueryFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static QueryFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** QueryFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static QueryFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** QueryFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
