     ///////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM    //
   // Written by Andrew Broad                ///
  //                                        ////
 // UnaryOpFrame.java (1998:10:27 11:37)   /////
////////////////////////////////////////////////
// UnaryOpFrame is a subclass of Frame.    /////
//                                         /////
// Methods of UnaryOpFrame:                /////
// - static initialise                     /////
// - UnaryOpFrame(UnaryOp,Frame)           /////
// - UnaryOpFrame(UnaryOp,Frame,Hashtable) /////
// - static buildQueuedFrames              /////
// - buildFrame                            /////
// - getFrameClass                         /////
// - getNumber             [INHERITED]     /////
// - setNumber             [INHERITED]     /////
// - getSlotFiller(Slot)   [INHERITED]     /////
// - getSlotFiller(String) [INHERITED]     /////
// - setSlotFiller(Slot)   [INHERITED]     /////
// - setSlotFiller(String) [INHERITED]     /////
// - toString              [INHERITED]     /////
// - print                 [INHERITED]     /////
// - generateFIL           [INHERITED]     ////
// - static writeClass                     ///
// - static readClass                      //
////////////////////////////////////////////

import java.io.*;
import java.util.*;
import express.*;

/*******************************************************************************
** The class `UnaryOpFrame', a subclass of <a href="Frame.html">Frame</a>,
** represents a frame internally as an exjava UnaryOp object - such frames
** should only be used to represent unary operations in EXPRESS models.
*******************************************************************************/

public class UnaryOpFrame extends Frame {
	private static FrameClass _frameClass;
	private static boolean _initialised= false;
	private UnaryOp _underlyingUnaryOp;
	private static Vector/*UnaryOpFrame*/ _toBeBuilt;
	private static Hashtable/*UnaryOp,UnaryOpFrame*/ _cache;
	private Frame _theEntityOrType;
	
	/**********************************************************************/
	// static UnaryOpFrame.initialise : Express -> void
	/***********************************************************************
	** Initialises the UnaryOpFrame class by setting the class variable
	** _frameClass after looking up the `unary_operation' entity in the
	** given EXPRESS (meta)model.
	** @param metamodel An EXPRESS metamodel, which is expected to contain
	**  the entity `unary_operation'.
	***********************************************************************/

	public static void initialise (Express metamodel) {
		Entity e= metamodel.findEntity("unary_operation");
		if (e != null) {
			_frameClass= new FrameClass(e,metamodel);
		} else {
			System.err.println("Fatal error: failed to find "
			+ "ENTITY unary_operation in \"EXPRESS/"
			+ "frame_classes.exp\"!");
			System.exit(1);
		}
	}
	
	/**********************************************************************/
	// UnaryOpFrame constructor : UnaryOp ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UnaryOp object
	** (it does so by copying the reference, which is maximally efficient
	** but the client should not then alter the UnaryOp directly, as that
	** would compromise the integrity of the frame).
	** @param instanceData An exjava UnaryOp object to wrap in this frame.
	** @param theEntity The entity which contains this unary operation.
 	** @exception FrameAlreadyExists If an UnaryOp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public UnaryOpFrame (UnaryOp instanceData, Frame theEntity)
	                         throws FrameAlreadyExists
	{
		this(instanceData,theEntity,new Hashtable());
	}
	
	/**********************************************************************/
	// UnaryOpFrame constructor : (UnaryOp x Hashtable) ->
	/***********************************************************************
	** Constructs this frame by plugging in the given exjava UnaryOp object
	** (it does so by copying the reference, which is maximally efficient,
	** but the client should not then alter the UnaryOp directly, as that
	** would compromise the integrity of the frame). This constructor allows
	** the unary operation frame to be tagged with extra slots, given in a
	** hash table.
	** @param instanceData An exjava UnaryOp object to wrap in this frame.
	** @param theEntity The entity which contains this unary operation.
	** @param moreInstanceData Extra slots to tag this unary operation
	**  frame with.
	** @exception FrameAlreadyExists If an UnaryOp has already been
	**  constructed for `instanceData'.
	***********************************************************************/

	public UnaryOpFrame (UnaryOp instanceData, Frame theEntity,
	                     Hashtable moreInstanceData)
	                    throws FrameAlreadyExists
	{
		super(moreInstanceData);

		if (_cache == null)
			_cache= new Hashtable();			

		_underlyingUnaryOp= instanceData;
		_theEntityOrType= theEntity;

		/////////////////////////////////////////////////////
		// Check that an UnaryOpFrame has not already been //
		// constructed for this UnaryOp.                   //
		/////////////////////////////////////////////////////		
		if (_cache.containsKey(_underlyingUnaryOp))
			throw new FrameAlreadyExists((Frame)_cache.
				get(_underlyingUnaryOp));

		_cache.put(_underlyingUnaryOp,this);

		////////////////////////////////////////////////////////////
		// Build the frame now, provided its frame class has been //
		// constructed, otherwise queue it to be built later.     //
		////////////////////////////////////////////////////////////
		if (_initialised) {
			buildFrame();
			_frameClass.validateFrame(this);
			try {
				_frameClass.addInstance(this);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone badly "
				+ "wrong for an InvalidFrameClass to be thrown "
				+ "here!");
			}
			setNumber(_frameClass.getInstances().size());
		} else {
			if (_toBeBuilt == null)
				_toBeBuilt= new Vector();
			_toBeBuilt.addElement(this);
		}
	}
	
	/**********************************************************************/
	// static UnaryOpFrame.buildQueuedFrames : void -> void
	/***********************************************************************
	** Builds any UnaryOp Frames that are queued to be built (those that
	** were constructed before the frame class was constructed). Their slots
	** are filled from the exjava objects with which they were constructed,
	** and they are then validated.
	***********************************************************************/

	public static void buildQueuedFrames () {
		_initialised= true;
		if (_toBeBuilt == null)
			return;

		Enumeration toBeBuilt= _toBeBuilt.elements();
		while (toBeBuilt.hasMoreElements()) {
			UnaryOpFrame e= (UnaryOpFrame)toBeBuilt.nextElement();
			e.buildFrame();
			_frameClass.validateFrame(e);
			try {
				_frameClass.addInstance(e);
			} catch (InvalidFrameClass x) {
				System.err.println("Something has gone "
				+ "badly wrong for an InvalidFrameClass"
				+ " to be thrown here!");
			}
			e.setNumber(_frameClass.getInstances().size());
		}
	}

	/**********************************************************************/
	// UnaryOpFrame.buildFrame : void -> void
	/***********************************************************************
	** Fills in the slots for this frame from the underlying exjava UnaryOp
	** object.
	***********************************************************************/

	public void buildFrame () {
		try {
			///////////////////////////////
			// operator : unary_operator //
			///////////////////////////////
			String enum= "";
			switch (_underlyingUnaryOp.kind) {
				case UnaryOp.NOT: enum= "not_operator"; break;
				case UnaryOp.NEGATE: enum= "unary_minus"; break;
			}
			setSlotFiller("operator", new EnumSymbol(enum));
	
			/////////////////////////////
			// expression : expression //
			/////////////////////////////
			setSlotFiller("expression", Values.convert
				(_underlyingUnaryOp.exp, _theEntityOrType));

		} catch (Exception x) {
			System.err.println("An exception of type "
			+ x.getClass() + " has been caught in "
			+ "UnaryOpFrame.buildFrame!");
		}
	}

	/**********************************************************************/
	// UnaryOpFrame.getFrameClass : void -> FrameClass
	/***********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	***********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}
	
	/**********************************************************************/
	// static UnaryOpFrame.writeClass : ObjectOutput -> void
	/***********************************************************************
	** Serialises the static data members (class variables) of
	** UnaryOpFrame.
	** @param oo An output object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClass);
		oo.writeObject(_cache);
	}

	/**********************************************************************/
	// static UnaryOpFrame.readClass : ObjectInput -> void
	/***********************************************************************
	** Deserialises the static data members (class variables) of
	** UnaryOpFrame.
	** @param ois An input object to save the class data to.
	** @exception IOException If an I/O exception occurs.
	***********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClass= (FrameClass)oi.readObject();
			_cache= (Hashtable)oi.readObject();
			_initialised= true;
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the "
			+ "stupid compiler is just too dumb to realise this!");
		}
	}
}
