     ////////////////////////////////////////
    // THE CONSTRAINT-UNDERSTANDING SYSTEM //
   // Written by Andrew Broad             ///
  //                                     ////
 // Values.java (1998:11:25 11:50)      /////
/////////////////////////////////////////////
// Methods of Values:                  /////
// - static convert(Object)            ////
// - static convert(Object,Frame)      ///
// - static FILise                     //
////////////////////////////////////////

import java.util.*;
import express.*;

/*******************************************************************************
** The class `Values' provides a library of utilities for values (values of slot
** fillers and values to be wrapped in defined type instances) such as
** validating them with respect to their type in the frame system, converting
** values from exjava objects to values for frame slot fillers, and writing the
** FIL for a value. This class is not meant to be instantiated.
*******************************************************************************/

public class Values {
	private Values () {}

	/**********************************************************************/
	// static Values.FILise : Object x boolean -> String
	/***********************************************************************
	** Writes FIL for a given value. If it is a frame, it generates a
	** corresponding hypertext link. If it is an aggregation, it sticks it
	** in brackets and recurses for each element. Otherwise, it just leaves
	** it as it is.
	** @param value The value to be FIL-ised.
	** @param HTML If true, FIL will be augmented with HTML tags.
	** @returns The FIL for `value'.
	***********************************************************************/

	public static String FILise (Object value, boolean HTML) {
		if (value instanceof Frame) {
			String returnValue= "";
			if (HTML) returnValue=
			          "<A HREF=\"" + ((Frame)value).getFrameClass()
			          +".html"+"#"+((Frame)value).getNumber()+"\">";
			returnValue+= ((Frame)value).getFrameClass() + "#" +
			              ((Frame)value).getNumber();
			if (HTML) returnValue+= "</A>";
			return returnValue;
		}
		if (value instanceof DefinedTypeInstance) {
			try {
				Frame type= ((DefinedTypeInstance)value).
				            getType().getTypeFrame();
				//////////////////
				// Enumeration? //
				//////////////////
				if (type.getSlotFiller("tipe") instanceof
				   DefinedTypeInstance &&
				   ((String)((DefinedTypeInstance)type.
				    getSlotFiller("tipe")).getType().
				    getTypeFrame().getSlotFiller("name")
				   ).equals("enumeration_type"))
				   	return type.getSlotFiller("name") + "("
				   	+ value + ")";
				else
					return type.getSlotFiller("name") + "("
					+ FILise(((DefinedTypeInstance)value).
					  getValue(),HTML) + ")";
			} catch (InvalidSlot x) {
				System.err.println("Something has gone badly " +
				"wrong for Values.FILise to throw an Invalid" +
				"Slot exception!");
				System.exit(1);
			}
		}
		if (value instanceof OrderedVector)
			return ((OrderedVector)value).FILise(HTML);
		if (value instanceof UnorderedVector)
			return ((UnorderedVector)value).FILise(HTML);
		if (value instanceof Logical) {
			switch (((Logical)value).value) {
				case Logical.TRUE: return "true";
				case Logical.FALSE: return "false";
				default: return "unknown";
			}
		}
		if (value instanceof String)
			return "\"" + value + "\"";
		if (value instanceof Variable)
			return "?" + value.toString();
		if (value instanceof Null)
			return "null";
		if (value == null)
			return null;
		else
			return value.toString();
	}

	/**********************************************************************/
	// static Values.convert : Object -> Object
	/***********************************************************************
	** Converts a given object from exjava to the corresponding object for
	** the frame system.
	** @param value An object from exjava.
	** @returns The corresponding object for the frame system. If it doesn't
	**  know to convert it, it returns it as is.
	***********************************************************************/

	public static Object convert (Object source) {
		try {
			if (source instanceof Type)
				return new DefinedTypeFrame((Type)source);
			if (source instanceof Entity)
				return new EntityFrame((Entity)source);
			if (source instanceof ExplicitAttr)
				return new ExplicitAttrFrame
					((ExplicitAttr)source);
			if (source instanceof InverseAttr)
				return new InverseAttrFrame
					((InverseAttr)source);
			if (source instanceof UniqueRule)
				return new UniqueRuleFrame((UniqueRule)source);
			if (source instanceof Literal)
				return new LiteralFrame((Literal)source);
			if (source instanceof BuiltIn)
				return new BuiltInFrame((BuiltIn)source);
			if (source instanceof SupExp)
				switch (((SupExp)source).kind) {
					case SupExp.ENTITY: return new
					                   EntityFrame(((SupExp)
					                   source).entity);
					case SupExp.ONEOF: return new OneofFrame
					                   ((SupExp)source);
					case SupExp.AND:   return new AndFrame
					                   ((SupExp)source);
					case SupExp.ANDOR: return new AndorFrame
					                   ((SupExp)source);
				}
			if (source instanceof AggType)
				return new AggTypeFrame((AggType)source);

			if (source instanceof NamedType)
				return convert(((NamedType)source).decl);

			if (source instanceof SimpleType) {
				String enum= "";
				switch (((SimpleType)source).kind) {
					case SimpleType.BOOLEAN:
						enum= "boolean_type"; break;
					case SimpleType.LOGICAL:
						enum= "logical_type"; break;
					case SimpleType.NUMBER:
						enum= "number_type";  break;
					case SimpleType.INTEGER:
						enum= "integer_type"; break;
					case SimpleType.REAL:
						enum= "real_type";    break;
					case SimpleType.STRING:
						enum= "string_type";  break;
					case SimpleType.BINARY:
						enum= "binary_type";  break;
				}
				return new DefinedTypeInstance
				(DefinedType.lookup("simple_type"),
				new EnumSymbol(enum));
			}
			if (source instanceof EnumType) {
				OrderedVector list= new OrderedVector
					(((EnumType)source).items);
				return new DefinedTypeInstance
				(DefinedType.lookup("enumeration_type"), list);
			}
			if (source instanceof SelectType) {
				UnorderedVector set= new UnorderedVector
					(((SelectType)source).items);
				return new DefinedTypeInstance
				(DefinedType.lookup("select_type"), set);
			}
		} catch (FrameAlreadyExists x) {
			return x._frame;
		} catch (Exception y) {
			System.err.println("An exception of type "
			+ y.getClass() + " has been caught in Values.convert");
		}
		return source;
	}

	/**********************************************************************/
	// static Values.convert : (Object x Frame) -> Object
	/***********************************************************************
	** Converts a given object from exjava to the corresponding object for
	** the frame system. To be used for converting IdRef, which requires
	** an entity frame to be passed in as a parameter, and for converting
	** anything that might contain an IdRef.
	** @param value An object from exjava.
	** @param entityOrTypeFrame A frame representing the current entity (in
	** the scope of which the IdRef is interpreted).
	** @returns The corresponding object for the frame system. If it doesn't
	**  know to convert it, it passes it on to convert(Object).
	***********************************************************************/

	public static Object convert (Object source, Frame entityOrTypeFrame) {
		if (source instanceof IdRef
		 && (entityOrTypeFrame instanceof EntityFrame
		  || entityOrTypeFrame instanceof DefinedTypeFrame))
		{
			Enumeration queryVariables= null;
			if (entityOrTypeFrame instanceof EntityFrame)
{
				queryVariables= ((EntityFrame)
				                 entityOrTypeFrame).
				                getQueryVariables().elements();
}
			else if (entityOrTypeFrame instanceof DefinedTypeFrame)
{
				queryVariables= ((DefinedTypeFrame)
				                 entityOrTypeFrame).
				                getQueryVariables().elements();
}
			else {
				System.err.println("Values.convert: " +
				"entityOrTypeFrame is not either an entity_type"
				+ " or a defined_type!");
				System.exit(1); return null;
			}
			while (queryVariables.hasMoreElements()) {
				try {
					Frame qVFrame;
					if ((qVFrame= (Frame)queryVariables.
					                     nextElement()).
					    getSlotFiller("name").
					    equals(((IdRef)source).id))
						return qVFrame;
				} catch (InvalidSlot x) {
					System.err.println("Slot \"query_" +
						"variables.name\" not found!");
					System.exit(1);
				}
			}
			try {
				return ((EntityFrame)entityOrTypeFrame).
				       lookupAttribute(((IdRef)source).id);
			} catch (InvalidAttribute x) {
//System.err.println("Warning: IdRef \"" + ((IdRef)source).id + "\" not resolved!");
				try {
					FrameClass frameClass= FrameClass.lookup
						("unresolved_reference");
					Frame frame= new Frame(frameClass,
					                 new Hashtable());
					frame.setSlotFiller("id",
					                    ((IdRef)source).id);
					return frame;
				} catch (FrameClassDoesNotExist xx) {
					System.err.println("Fatal error: frame "
					+ "class \"unresolved_reference\" does"
					+ " not exist!");
					System.exit(1);
				} catch (InvalidSlot yy) {
					System.err.println("Fatal error: slot "
					+ "\"unresolved_reference:id\" does"
					+ " not exist!");
					System.exit(1);
				}
				return null;
			}
		}
		if (source instanceof IdRef
		 && ! (entityOrTypeFrame instanceof EntityFrame
		    || entityOrTypeFrame instanceof DefinedTypeFrame))
		{
			System.err.println("Values.convert(Object,Frame) "
			+ "doesn't yet know to handle IdRefs when the second "
			+ "argument is anything other than an EntityFrame or a "
			+ "DefinedTypeFrame!");
			System.exit(1);
		}
		if (source instanceof BinaryOp) {
			try {
				return new BinaryOpFrame((BinaryOp)source,
			                                 entityOrTypeFrame);
			} catch (FrameAlreadyExists a) {
				return a._frame;
			}
		}
		if (source instanceof UnaryOp) {
			try {
				return new UnaryOpFrame((UnaryOp)source,
			                               entityOrTypeFrame);
			} catch (FrameAlreadyExists b) {
				return b._frame;
			}
		}
		if (source instanceof Query) {
			try {
				return new QueryFrame((Query)source,
			                               entityOrTypeFrame);
			} catch (FrameAlreadyExists c) {
				return c._frame;
			}
		}
		if (source instanceof DomainRule) {
			try {
				return new DomainRuleFrame((DomainRule)source,
			                           entityOrTypeFrame);
			} catch (FrameAlreadyExists d) {
				return d._frame;
			}
		}
		if (source instanceof BuiltInFnCall) {
			try {
				return new BuiltInFnCallFrame((BuiltInFnCall)
				                              source,
			                                     entityOrTypeFrame);
			} catch (FrameAlreadyExists e) {
				return e._frame;
			}
		}
		if (source instanceof UserFnCall) {
			try {
				return new UserFnCallFrame((UserFnCall)source,
			                                  entityOrTypeFrame);
			} catch (FrameAlreadyExists f) {
				return f._frame;
			}
		}
		if (source instanceof Qualifier) {
			try {
				return new QualifierFrame((Qualifier)source,
				                          entityOrTypeFrame);
			} catch (FrameAlreadyExists g) {
				return g._frame;
			}
		}
		if (source instanceof AggInit) {
			try {
				return new AggInitFrame((AggInit)source,
				                        entityOrTypeFrame);
			} catch (FrameAlreadyExists h) {
				return h._frame;
			}
		}
		if (source instanceof AggInitElem) {
			try {
				return new AggInitElemFrame((AggInitElem)source,
				                            entityOrTypeFrame);
			} catch (FrameAlreadyExists i) {
				return i._frame;
			}
		}
		if (source instanceof DerivedAttr) {
			try {
				return new DerivedAttrFrame((DerivedAttr)source,
				                            entityOrTypeFrame);
			} catch (FrameAlreadyExists j) {
				return j._frame;
			}
		}
		return convert(source);
	}
}